<?php

/**
 * A library that uses the cybertill api to make orders and sync stock
 * SOAP API docs: http://ct28790.c-pos.co.uk/current/docs/soap_documentation_v1_5/index.html
 */

class Cybertill {

	//The current loaded config (live/dev)
	protected $_config = null;

	//The current Soap Client
	protected $_client = null;

	//The current authorisation id returned from the client
	protected $_authId = null;

	//The current customer id, taken from config if avilable created from customer details
	protected $_customer = null;

/**
 * Make an order on the client's cybertill system.
 * @param  array 		$order An EvCheckout order array consisting of order info, order items, order totals and order data.
 * @return array|bool   Returns transaction array. Contains success/failure and transaction info/errors.
 */
	public function makeOrder($order) {
		if ($this->_setupClient()) {
			$this->_setupCustomer();

			if (!empty($this->_customer)) {
				$cybertillOrderDetails = $this->_getDetailsForOrder($order);

				$cybertillOrderItems = $this->_getItemsForOrder($order);

				$cybertillOrderDelivery = $this->_getDeliveryForOrder($order);

				$cybertillOrderPayments = $this->_getPaymentsForOrder($order);

				if (
					!empty($cybertillOrderDetails) &&
					!empty($cybertillOrderItems) &&
					!empty($cybertillOrderDelivery) &&
					!empty($cybertillOrderPayments)
				) {
					$transaction = $this->_client->transaction_add_giftcard(
						$cybertillOrderDetails,
						$cybertillOrderItems,
						$cybertillOrderDelivery,
						$cybertillOrderPayments
					);

					if ($transaction->success) {
						if (!empty($this->_config['despatch_user_id'])) {
							$despatchOrder = [
								[
									'transactionId' => $transaction->transaction->id,
									'locationId' => $this->_config['location_id'],
									'consignmentRef' => $this->_config['order_details']['order_note'],
									'userId' => null
								]
							];

							$dispatched = $this->_client->transaction_despatch_orders(
								$this->_config['despatch_user_id'],
								$despatchOrder
							);

							if ($despatchOrder->success) {
								return true;
							} else {
								if (CakePlugin::loaded('EvErrbit')) {
									$e = new Exception(print_r($transaction));
									Errbit::notify($e);
								}
								return false;
							}
						}
					} else {
						if (CakePlugin::loaded('EvErrbit')) {
							$e = new Exception(print_r($transaction));
							Errbit::notify($e);
						}
					}
				}
			}
		}

		return false;
	}

/**
 * Sync the stock between the website and cybertill. It doesn't update cybertill information.
 * @return bool Returns true if the stock has attempted to be updated. False if no items were
 *                      found to be synchronised.
 */
	public function syncStock() {
		if ($this->_setupClient()) {
			if (!empty($this->_config['models'])) {
				$models = $this->_config['models'];
				if (!empty($models['item_model']) && !empty($models['stock_model'])) {
					$ItemModel = EvClassRegistry::init($models['item_model']);

					$items = $ItemModel->find('all');

					foreach ($items as $item) {
						if (!empty($item[$ItemModel->alias]['sku'])) {
							$cybertillItem = $this->_getItem($item[$ItemModel->alias]['sku']);

							if (!empty($cybertillItem)) {
								$stock = $this->_getItemStock($cybertillItem->productOption->id, $this->_config['location_id']);
								if (!empty($stock)) {
									$this->_updateStock($stock, $item, $ItemModel);
								}
							}
						}
					}

					return true;
				}
			}
		}

		return false;
	}

/**
 * Check the stock of a particular item by using the SKU provided.
 * @param  string   $sku The SKU to be used to find the item to return the stock of
 * @return int|bool      If an item and stock are found then stock will be returned otherwise false is returned
 */
	public function checkStock($sku) {
		if (empty($this->_client)) {
			$this->_setupClient();
		}

		if (!empty($this->_client)) {
			$productItem = $this->_getItem($sku);

			if (!empty($productItem)) {
				return $this->_getItemStock($sku, $productItem->productOption->id, $this->_config['location_id']);
			}
		}

		return false;
	}

/**
 * Create the soap client to create orders or sync stock. Uses settings from the config, if a requried
 * config setting is missing then the connection won't be made.
 */
	protected function _setupClient() {
		$this->_getConfig();

		if (!empty($this->_config['client_details'])) {
			$clientDetails = $this->_config['client_details'];
			$this->_client = new SoapClient($clientDetails['wsdl']);

			if (!empty($clientDetails['url']) && !empty($clientDetails['user_id'])) {
				$this->_authId = $this->_client->authenticate_get($clientDetails['url'], $clientDetails['user_id']);

				$this->_client = new SoapClient($clientDetails['wsdl'], ['login' => $this->_authId]);

				return true;
			}
		}

		// A client hasn't been able to be made
		if (CakePlugin::loaded('EvErrbit')) {
			$e = new Exception("A client wasn't able to be established.");
			Errbit::notify($e);
		}

		return false;
	}

/**
 * Read the config settings based on the current app environment. Allows for different testing settings if available.
 */
	protected function _getConfig() {
		if (Configure::read('app.environment') == 'PRODUCTION') {
			$this->_config = Configure::read('EvCybertill.live');
		} else {
			$this->_config = Configure::read('EvCybertill.dev');
		}
	}

/**
 * Attempt to get a customer using the provided customer id in the config. If an id is missing in the config
 * then attempt to create a customer using the customer details in the config. If a customer is found or
 * added then save the details
 */
	protected function _setupCustomer() {
		if (!empty($this->_config['customer_id'])) {
			$this->_customer = $this->_client->customer_get($this->_config['customer_id']);
		} else {
			if (!empty($this->_config['customer_details'])) {
				$this->_customer = $this->_client->customer_add($this->_config['customer_details']);
			}
		}
	}

/**
 * Get all the cybertill items from the provided order. Each item is checked if it exists and is available before being
 * added to the returned array. If an item is missing then the order items aren't returned and the cybertill
 * transaction won't take place.
 * @param  array  $order An EvCheckout order
 * @return array         The cybertill order items or null if an item can't be found
 */
	protected function _getItemsForOrder($order) {
		$paths = $this->_config['paths'];

		if (!empty($paths['items'])) {
			$items = Hash::get($order, $paths['items']);

			foreach ($items as $item) {
				$orderItem = $this->_getItemForOrder($item, $paths['sku']);
				if (!empty($orderItem)) {
					$orderItems[] = $orderItem;
				}
			}

			if (!empty($orderItems) && count($orderItems) == count($items)) {
				return $orderItems;
			}
		}

		return null;
	}

/**
 * Get an item for a cybertill order. Check if the item has stock available before trying to order it.
 * An cybertill order item array is constructed if the item is found and has stock and returned to add
 * to the order items array.
 * @param  array  $orderItem An item provided by EvCheckout to add to the cybertill order
 * @param  string $skuPath   The path within the orderItem array where the sku is stored.
 * @return array             The cybertill order item or null if no item was found.
 */
	protected function _getItemForOrder($orderItem, $skuPath) {
		// Get the product the client wants us to add, we need this for the product ID.
		$sku = Hash::get($orderItem, $skuPath);

		// Check the item is in stock before we try and add it to an order.
		$stock = $this->checkStock($sku);
		$productItem = $this->_getItem($sku);

		if (!empty($stock) && !empty($productItem)) {
			return [
				'itemId' => $productItem->productOption->id,
				'salesQty' => $orderItem['quantity'],
				'itemPrice' => $orderItem['unit_price_inc_tax'],
				'note' => '',
				'vatRate' => $orderItem['tax_rate']
			];
		}

		return null;
	}

/**
 * Get an item from cybertill using the SKU provided.
 * @param  string $sku The SKU of the item to get from cybertill
 * @return object      The cybertill item or null if no item was found
 */
	protected function _getItem($sku) {
		try {
			return $this->_client->item_get(null, $sku, null, true);
		} catch (Exception $e) {
			if (CakePlugin::loaded('EvErrbit')) {
				$e = new Exception("no product exists with sku: " . $sku);
				Errbit::notify($e);
			} else {
				CakeLog::write('error', "no product exists with sku: " . $sku);
			}

			return null;
		}
	}

/**
 * Get the stock for a specific cybertill item. Items are found via their SKU's.
 * It is possible that if the item has been discountinued that an item would be found but no stock
 * exists for it. In this case then an item won't be returned
 * @param  string $sku        The SKU of the item to get stock of
 * @param  int    $itemId     The cybertill if of the item to get stock of
 * @return int                The current stock in cybertill of the item to get stock of
 */
	protected function _getItemStock($sku, $itemId) {
		$stock = null;
		if (!empty($itemId)) {
			try {
				$stock = $this->_client->stock_item($itemId, $this->_config['location_id']);
			} catch (Exception $e) {
				if (CakePlugin::loaded('EvErrbit')) {
					$e = new Exception("no stock found for: " . $sku);
					Errbit::notify($e);
				} else {
					CakeLog::write('error', "no stock found for: " . $sku);
				}

				// Set product to inactive if no stock is found as it has been discontinued
			}
		}
		return $stock;
	}

/**
 * Get the order details for a cybertill order. The details are mainly taken from the config but the order total
 * is taken from the order created on the webite.
 * @param  array $order An array containing the order to get the order total from
 * @return array        The array containing the order details
 */
	protected function _getDetailsForOrder($order) {
		$orderDetails = $this->_config['order_details'];

		if (!empty($orderDetails)) {
			if (!empty($orderDetails['order_total_label'])) {
				//Get the order total from the order
				$orderDetails['order_total'] = $this->_getOrderTotalForOrder($order, $orderDetails['order_total_label']);

				if (!empty($orderDetails['order_total'])) {
					return [
						'websiteId' => $orderDetails['website_id'],
						'customerId' => $this->_customer->customer->id,
						'locationId' => $this->_config['location_id'],
						'orderTotal' => $orderDetails['order_total'],
						'orderNote' => $orderDetails['order_note'],
						'customerOrderRef' => $orderDetails['customer_order_reference'],
					];
				}
			}
		}
	}

/**
 * Get a total of an order to be used in the payment details and order details of the cybertill
 * order. The specific total to get is based on the label passed through which itself is in the config.
 * @param  array  $order An array containing the order to get the order total from
 * @param  string $label The lable of the total to return
 * @return float         The amount of total. If a tax inclusive price is available then it will return that.
 */
	protected function _getOrderTotalForOrder($order, $label) {
		$orderTotal = null;

		if (!empty($this->_config['paths']['totals'])) {
			$totals = Hash::get($order, $this->_config['paths']['totals']);

			foreach ($totals as $total) {
				if ($total['name'] == $label) {
					if (!empty($total['display_inc_tax'])) {
						$orderTotal = $total['display_inc_tax'];
					} else {
						$orderTotal = $total['amount'];
					}
				}
			}
		}

		return $orderTotal;
	}

/**
 * Get the delivery details for a cybertill order. The details are based on the user and as we use one user for all orders
 * we can't change it based on the actual delivery address of the order.
 * @param  array  $order An array containing the order to get the delivery details from
 * @return array         The array containing the payment details
 */
	protected function _getDeliveryForOrder($order) {
		return [
			'isCollection' => true,
			'addressId' => $this->_customer->addresses->item->id,
			'serviceId' => null,
			'tariffId' => null,
			'tariff' => 0,
			'vatRate' => null,
			'recipient' => (!empty($this->_config['order_delivery']['recipient'])) ? $this->_config['order_delivery']['recipient'] : '',
			'dateRequired' => '',
			'when' => null,
			'instructions' => (!empty($this->_config['order_delivery']['instructions'])) ? $this->_config['order_delivery']['instructions'] : '',
			'giftMessage' => (!empty($this->_config['order_delivery']['giftMessage'])) ? $this->_config['order_delivery']['giftMessage'] : '',
			'giftReceipt' => false
		];
	}

/**
 * Get the payment details for a cybertill order. By default the payment is recorded as a card payment unless
 * specified otherwise in the config.
 * @param  array  $order An array containing the order to get the payment details from
 * @return array         The array containing the payment details
 */
	protected function _getPaymentsForOrder($order) {
		$paymentDetails = $this->_config['order_payments'];

		if (!empty($paymentDetails['total_label'])) {
			//Get the payment total from the order
			$orderDetails['total'] = $this->_getOrderTotalForOrder($order, $paymentDetails['total_label']);

			return [
				[
					'type' => (!empty($this->_config['order_payments']['type'])) ? $this->_config['order_payments']['type'] : 2,
					'total' => $orderDetails['total']
				]
			];
		}
	}

/**
 * Update the stock in the website database using the stock from cybertill
 * @param  int 	  $cybertillStock The amount of stock currently in cybertill
 * @param  array  $item           The item to update
 * @param  obj    &$ItemModel     The Model of the item to update
 * @return bool                   Whether the update was successful or not
 */
	protected function _updateStock($cybertillStock, $item, &$ItemModel) {
		// Update the inventory with the stock from cybertill
		$StockModel = EvClassRegistry::init($this->_config['models']['stock_model']);

		$stock = $StockModel->find(
			'first',
			[
				'conditions' => [
					'id' => $item['Inventory']['id']
				]
			]
		);

		//Item exists on site so we can update inventory
		$stock[$StockModel->alias]['stock'] = $cybertillStock->item->stock;

		return $StockModel->save($stock);
	}
}
